<?php
/*
readout a Fritz AHA Device (Fritz Powerline 546E, Fritz Dect200) from Fritzbox
(need FritzOS6.0 or 5.53+Labor Oct2013)
www.tdressler.net/ipsymcon
V0.4 12.10.2013
see also
http://www.avm.de/de/Extern/files/session_id/AVM_Technical_Note_-_Session_ID.pdf
http://www.avm.de/de/Extern/files/session_id/AHA-HTTP-Interface.pdf
Json Wrapper:
http://www.tdressler.net/ipsymcon/jsonapi.html

*/
if (!function_exists('mb_convert_encoding')) {
   die ("Need mbstring_extension!\n");
}
if (!function_exists('IPS_GetKernelVersion')) {
   //we are outside of IPS and need the JSON wrapper
   include('ips_wrapper.php');
}
/* config section, pls adopt this for your device */
$catname='FRITZ Smarthome';
$sensorname=''; //will be expanded with actor name
$host='fritzbox';
$password = 'Admin65'; //login password
$GLOBALS['autocreate']=true;
/* end config section */

//api urls
$loginurl="http://".$host."/login_sid.lua";//login_sid.lua for fritzos 5.50+
$ahaurl="http://".$host."/webservices/homeautoswitch.lua";//api url
//Variable definitions
$vartypes=array('ID'=>array('type'=>3,'profile'=>''),
               	'AIN'=>array('type'=>3,'profile'=>''),
               	'Host'=>array('type'=>3,'profile'=>''),
			   'Status'=>array('type'=>0,'profile'=>'~Switch'),
			   'Total Energy'=>array('type'=>2,'profile'=>'~Electricity'),
			   'Last Energy'=>array('type'=>2,'profile'=>'~Electricity'),
			   'Actual Power'=>array('type'=>2,'profile'=>'Power_W.3')
			);

//define profile for actual usage (getswitchpower) value
check_profile('Power_W.3',2,null,' W',0,null,1,3,false);

//login to switch
$sid=get_sid($loginurl,$password);
if (!isset($sid)) {
   IPS_LogMessage("FritzAHA Reader", "Login failed");
   return;

}

$actors=get_actors($ahaurl,$sid);
//retrieve
foreach ($actors as $ain) {
   $url=$ahaurl.'?sid='.$sid.'&ain='.$ain;
   //query device values
   $name = chop(file_get_contents($url."&switchcmd=getswitchname"));
   if (!$name) {
     IPS_LogMessage("FritzAHA Reader", "No Answer for name query on $host");
	  continue;
   }
   //get IPS Variable IDS and Values below defined section
   $varids=get_ips_vars($name,$vartypes,$catname,$sensorname);
   if (is_null($varids)) {
	  //no vars available, maybe autocreate disabled
	  IPS_LogMessage("FritzAHA Reader", "get_ips_vars failed");
	  continue;
   }
   $total=$varids['Total Energy']['val'];
   /* get_ain */
   $status = chop(file_get_contents($url."&switchcmd=getswitchstate"));
   $power = chop(file_get_contents($url."&switchcmd=getswitchpower"));
   $new_total = chop(file_get_contents($url."&switchcmd=getswitchenergy"));
   //prepare output
   if (preg_match("/^\d+$/",$power)) {
	  $actual=$power/1000; //mW->W
   }else{
	  $actual="?";
	  print "Error:".$name." getswitchpower=".$power;
	  continue;
   }
   if (preg_match("/^\d+$/",$new_total)) {
      $new_total=$new_total/1000; //Wh ->KWh
      $energy=$new_total-$total;
   }else{
      $energy="?";
      print "Error:".$name." getswitchenergy=".$new_total;
      continue;
   }

   //store in IPS
   SetValue($varids['Total Energy']['id'],$new_total);
   SetValue($varids['Last Energy']['id'],$energy);
   SetValue($varids['Actual Power']['id'],$actual);
   if ($varids['Status']['val'] != $status)SetValue($varids['Status']['id'],$status);
   if ($varids['AIN']['val'] != $ain) SetValue($varids['AIN']['id'],$ain);
   if (($varids['Host']['val']==0) || ($varids['Host']['val'] != $host)) SetValue($varids['Host']['id'],$host);

   //logging
   IPS_LogMessage("FritzAHA Reader", $name.":".(($status=="1")?"On":"Off")." (Act:".$actual." W, Diff: ".$energy." Wh Power: $power mW, Total:$new_total Kwh)");


}

/* end */
//-------------------------------------------------------------------------------
 /**
  * login into fritz and get sid
  * @param string loginurl url for accessing device
  * @param string password password for login
  * @return string sid
  */
function get_sid ($loginurl,$password) {

 // get challenge string
 $http_response = file_get_contents($loginurl);
 if(preg_match("/<Challenge>(\w+)<\/Challenge>/i",$http_response, $res)) {
   $challenge=$res[1];
 }
 //get SID
 if(preg_match("/<SID>([\da-f]+)<\/SID>/i",$http_response, $res)) {
   $sid=$res[1];
 }

 if ($sid && (preg_match("/^[0]+$/",$sid)) && $challenge) {
     //sid is null, got challenge
    $sid="";
    //build password response
    $pass=$challenge."-".$password;
    //UTF-16LE encoding as required
    $pass=mb_convert_encoding($pass, "UTF-16LE");
    //md5hash on top
    $md5 = md5($pass);
    //final answer string
    $challenge_response = $challenge."-".$md5;
    //send to box
    $url=$loginurl."?response=".$challenge_response;
    $http_response = file_get_contents($url);
    //check answer
    if(preg_match("/<SID>([\da-f]+)<\/SID>/i",$http_response, $res)) {
         //got answer with sid
         $sid=$res[1];
         if (!preg_match("/^[0]+$/",$sid)) {
            //is not null, bingo!
            return $sid;
        }
    }

 }
 return null;
}//function

//-------------------------------------------------------------------------------
/**
 * retrieve actor AINs from smarthome overview page
 * @param string url aha url
 * @param string sid SID to be used in query
 * @return array of string AINs
 */
function get_actors( $url,$sid) {
   $actors=array();
	$actorlist=chop(file_get_contents($url."?switchcmd=getswitchlist&sid=".$sid));
	$a=explode(',',$actorlist);
	foreach ($a as $actor)
	if (preg_match('/^(([a-fA-F0-9]{2}[:|\-]?){6})$/',$actor)) {
	   //AIN is 6x2Hex
		array_push($actors,$actor);
	}
	return $actors;
}
//-------------------------------------------------------------------------------
/**
* IPS Variablen handler
* creates variables as needed
* returns assoc. Array with IPS Variable ID and Value
* @param integer Sensor Address
* @param array Array with Variable Names, Types and Profiles
* @param string Sensor Group Master Categorie Name
* @param string Sensor default name, will be extended with $addr
*/
function get_ips_vars($addr,$vartypes,$cat,$sens) {

   $varids=null;
   $master=@IPS_GetObjectIDByName($cat,0);
   //no master cat, create new
   if (!$master) {
      $master=IPS_CreateCategory();
      IPS_SetName($master,$cat);
      IPS_SetParent($master,0);
      if ($master>0) {
         IPS_LogMessage('CUL', "Master category created, ID=$master\n");
      }else{
         IPS_LogMessage('CUL', "Can't create Master Category\n");
	return null;
      }
   }

  $id=0;
  if ($master>0) {
      //get chilren sensors
      $Sensors=IPS_GetChildrenIDs($master);
      foreach($Sensors as $sid) {
         $name=IPS_GetName($sid);
         $vars=IPS_GetChildrenIDs($sid);
         foreach($vars as $vid) {
            $obj=IPS_GetObject($vid);
            $vname=$obj['ObjectName'];
            $typ=$obj['ObjectType'];
            if ($typ==2) { //Variable
            //if ID, here is the address
               if ($vname="ID") {
                  $i=GetValue($vid);
                  //go out if matches, $id returns the sensor categorie id
                  if ($i===$addr) {
                     $id=$sid;
                     break;
                  }
               }
            }
         }
         if ($id>0) break;
      }
      if ($id==0) {
   	//Sensor with address $addr not found in IPS
         if ($GLOBALS['autocreate']==false) {
         //autocreate disable, ignore new device
            return null;
         }
         //create new sensor
         $id=ips_createCategory();
         ips_setName($id,$sens.' '.$addr);
         ips_setParent($id,$master);
         //creates all needed variables for the new sensor
         foreach (array_keys($vartypes) as $name) {
            $typ=$vartypes[$name]['type'];
            $profile=$vartypes[$name]['profile'];
            $vid=IPS_CreateVariable($typ);
            ips_setname($vid,$name);
            ips_setParent($vid,$id);
            IPS_SetVariableCustomProfile($vid,$profile);
            //preload variables
            SetValue($vid,0);
            $varids[$name]['id']=$vid;
            $varids[$name]['val']=0;
            //Store address in $ID for next time
            if ($name=='ID') {
               SetValue($vid,$addr);
               $varids[$name]['val']=$addr;
            }
        }
	}else{
         //found matching cat, collect ids and vals for this sensor
		$vars=IPS_GetChildrenIDs($id);
		foreach($vars as $vid) {
            $obj=IPS_GetObject($vid);
            $name=$obj['ObjectName'];
            $typ=$obj['ObjectType'];
            if ($typ==2) { //Variable
               $val=GetValue($vid);
               $varids[$name]['id']=$vid;
               $varids[$name]['val']=$val;
            }
		}
	}
    //returns IDs and Values of this Sensor, Name is Key
	return $varids;
   }
}
//-------------------------------------------------------------------------------
/**
* Get ID by name if exists, else create
* @Param String Name
* @param integer Parent-ID
* @return integer ID
*/
function getVid($name,$par) {
	$vid = @IPS_GetVariableIDByName($name, $par);
	if ($vid === false) {
		$vid=IPS_CreateVariable(3);
		ips_setname($vid,$name);
		ips_setParent($vid,$par);
		setValue($vid,'');
	}
	return $vid;
}
//-------------------------------------------------------------------------------
/**
* Check profile by name if exists, else create
* @Param String Name
* @param integer Variable Typ (0..3)
* @param String prefix befor value
* @param String suffix after value
* @param integer min value
* @param integer max value
* @param integer step value
* @param integer digits for formatting
* @param boolean drop existing profile first
*/

function check_profile($pname,$typ,$prefix,$suffix,$min,$max,$step,$digit=0,$drop=false) {
	if (IPS_VariableProfileExists($pname) && $drop) IPS_DeleteVariableProfile($pname);
	if (!IPS_VariableProfileExists($pname)) {
		if (IPS_CreateVariableProfile($pname,$typ)) {
		   	IPS_SetVariableProfileText($pname,$prefix,$suffix);
			if (isset($min) && isset($max) && isset($step)) {
		   		IPS_SetVariableProfileValues($pname,$min,$max,$step);
		   	}
	  		if (isset($digit)) {
		   		IPS_SetVariableProfileDigits($pname,$digit);
		   	}
 		}
	}
}
?>